//=======================================================================================
// ____          ____    __    ______     __________   __      __       __        __         
// \    \       |    |  |  |  |   _   \  |___    ___| |  |    |  |     /  \      |  |        
//  \    \      |    |  |  |  |  |_)   |     |  |     |  |    |  |    /    \     |  |        
//   \    \     |    |  |  |  |   _   /      |  |     |  |    |  |   /  /\  \    |  |        
//    \    \    |    |  |  |  |  | \  \      |  |     |   \__/   |  /  ____  \   |  |____    
//     \    \   |    |  |__|  |__|  \__\     |__|      \________/  /__/    \__\  |_______|   
//      \    \  |    |   ________________________________________________________________    
//       \    \ |    |  |  ______________________________________________________________|   
//        \    \|    |  |  |         __          __     __     __     ______      _______    
//         \         |  |  |_____   |  |        |  |   |  |   |  |   |   _  \    /  _____)   
//          \        |  |   _____|  |  |        |  |   |  |   |  |   |  | \  \   \_______    
//           \       |  |  |        |  |_____   |   \_/   |   |  |   |  |_/  /    _____  |
//            \ _____|  |__|        |________|   \_______/    |__|   |______/    (_______/   
//
//  This file is part of VirtualFluids. VirtualFluids is free software: you can 
//  redistribute it and/or modify it under the terms of the GNU General Public
//  License as published by the Free Software Foundation, either version 3 of 
//  the License, or (at your option) any later version.
//  
//  VirtualFluids is distributed in the hope that it will be useful, but WITHOUT 
//  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
//  for more details.
//  
//  You should have received a copy of the GNU General Public License along
//  with VirtualFluids (see COPYING.txt). If not, see <http://www.gnu.org/licenses/>.
//
//=======================================================================================
#ifndef STREAM_MANAGER_H
#define STREAM_MANAGER_H

#include <vector>
#include "Core/DataTypes.h"

#include <cuda_runtime.h>

class CudaStreamManager
{
private:
    std::vector<cudaStream_t> cudaStreams;
    cudaEvent_t startBulkKernel = NULL;
    const int precursorStreamIndex    = 2;
    const int borderStreamIndex       = 1;
    const int bulkStreamIndex         = 0;

public:
    void launchStreams(uint numberOfStreams);
    void terminateStreams();
    cudaStream_t &getStream(uint streamIndex);
    int getBorderStreamIndex();
    int getBulkStreamIndex();
    int getPrecursorStreamIndex();

    // Events
    void createCudaEvents();
    void destroyCudaEvents();
    void triggerStartBulkKernel(int streamIndex);
    void waitOnStartBulkKernelEvent(int strteamIndex);
};

#endif
